""" Common code for push to GitHub or GitLab """

import abc
import argparse
import importlib
import re
from typing import cast, Iterable, Optional

from path import Path
import ui

import tsrc
import tsrc.git


def service_from_url(url: str) -> str:
    """
    >>> service_from_url("git@github.com:foo/bar")
    'github'
    >>> service_from_url("git@gitlab.local:foo/bar")
    'gitlab'

    """
    if url.startswith("git@github.com"):
        return "github"
    else:
        return "gitlab"


def project_name_from_url(url: str) -> str:
    """
    >>> project_name_from_url('git@example.com:foo/bar.git')
    'foo/bar'
    >>> project_name_from_url('ssh://git@example.com:8022/foo/bar.git')
    'foo/bar'

    """
    # split everthing that is separated by a colon or a slash
    parts = re.split("[:/]", url)
    # join the last two parts
    res = "/".join(parts[-2:])
    # remove last `.git`
    if res.endswith(".git"):
        res = res[:-4]
    return res


class RepositoryInfo:
    def __init__(self, working_path: Path = None) -> None:
        self.project_name = None  # type: Optional[str]
        self.url = None  # type: Optional[str]
        self.path = None  # type: Optional[Path]
        self.current_branch = None  # type: Optional[str]
        self.service = None  # type: Optional[str]
        self.tracking_ref = None  # type: Optional[str]
        self.read_working_path(working_path=working_path)

    def read_working_path(self, working_path: Path = None) -> None:
        self.path = tsrc.git.get_repo_root(working_path=working_path)
        self.current_branch = tsrc.git.get_current_branch(self.path)
        rc, out = tsrc.git.run_captured(self.path, "remote", "get-url", "origin", check=False)
        if rc == 0:
            self.url = out
        if not self.url:
            return
        self.tracking_ref = tsrc.git.get_tracking_ref(self.path)
        self.project_name = project_name_from_url(self.url)
        self.service = service_from_url(self.url)


class PushAction(metaclass=abc.ABCMeta):
    def __init__(self, repository_info: RepositoryInfo, args: argparse.Namespace) -> None:
        self.args = args
        self.repository_info = repository_info

    @property
    def repo_path(self) -> Path:
        return self.repository_info.path

    @property
    def tracking_ref(self) -> Optional[str]:
        return self.repository_info.tracking_ref

    @property
    def remote_name(self) -> Optional[str]:
        if not self.tracking_ref:
            return None
        return self.tracking_ref.split("/", maxsplit=1)[0]

    @property
    def current_branch(self) -> Optional[str]:
        return self.repository_info.current_branch

    @property
    def remote_branch(self) -> Optional[str]:
        if not self.tracking_ref:
            return self.current_branch
        else:
            return self.tracking_ref.split("/", maxsplit=1)[1]

    @property
    def requested_target_branch(self) -> Optional[str]:
        return cast(Optional[str], self.args.target_branch)

    @property
    def requested_title(self) -> Optional[str]:
        return cast(Optional[str], self.args.title)

    @property
    def requested_reviewers(self) -> Iterable[str]:
        return cast(Iterable[str], self.args.reviewers)

    @property
    def requested_assignee(self) -> Optional[str]:
        return cast(Optional[str], self.args.assignee)

    @property
    def project_name(self) -> Optional[str]:
        return self.repository_info.project_name

    @abc.abstractmethod
    def setup_service(self) -> None:
        pass

    @abc.abstractmethod
    def post_push(self) -> None:
        pass

    def push(self) -> None:
        ui.info_2("Running git push")
        remote_name = self.remote_name or "origin"
        if self.args.push_spec:
            push_spec = self.args.push_spec
        else:
            push_spec = "%s:%s" % (self.current_branch, self.remote_branch)
        cmd = ["push", "-u", remote_name, push_spec]
        if self.args.force:
            cmd.append("--force")
        tsrc.git.run(self.repo_path, *cmd)

    def execute(self) -> None:
        self.setup_service()
        self.push()
        self.post_push()


def main(args: argparse.Namespace) -> None:
    repository_info = RepositoryInfo()
    service_name = repository_info.service
    module = importlib.import_module("tsrc.cli.push_%s" % service_name)
    push_action = module.PushAction(repository_info, args)  # type: ignore
    push_action.execute()



""" Parse tsrc config files """

from path import Path
import ruamel.yaml
import schema
from typing import Any, Dict, NewType
import xdg

import tsrc

Config = NewType('Config', Dict[str, Any])


def parse_config_file(
        file_path: Path, config_schema: schema.Schema, roundtrip: bool = False) -> Config:
    try:
        contents = file_path.text()
    except OSError as os_error:
        raise tsrc.InvalidConfig(file_path, os_error)
    try:
        if roundtrip:
            parsed = ruamel.yaml.safe_load(contents, ruamel.yaml.RoundTripLoader)
        else:
            parsed = ruamel.yaml.safe_load(contents)
    except ruamel.yaml.error.YAMLError as yaml_error:
        raise tsrc.InvalidConfig(file_path, yaml_error)
    try:
        config_schema.validate(parsed)
    except schema.SchemaError as schema_error:
        raise tsrc.InvalidConfig(file_path, schema_error)
    return Config(parsed)


def get_tsrc_config_path() -> Path:
    config_path = Path(xdg.XDG_CONFIG_HOME)
    config_path = config_path / "tsrc.yml"
    return config_path


def dump_tsrc_config(config: Config) -> None:
    dumped = ruamel.yaml.dump(config, Dumper=ruamel.yaml.RoundTripDumper)
    file_path = get_tsrc_config_path()
    file_path.write_text(dumped)


def parse_tsrc_config(config_path: Path = None, roundtrip: bool = False) -> Config:
    auth_schema = {
        schema.Optional("gitlab"): {"token": str},
        schema.Optional("github"): {"token": str},
    }
    tsrc_schema = schema.Schema({"auth": auth_schema})
    if not config_path:
        config_path = get_tsrc_config_path()
    return parse_config_file(config_path, tsrc_schema, roundtrip=roundtrip)



from typing import Any
import os
from path import Path
import pytest

import tsrc.cli.main


class CLI():
    def __init__(self) -> None:
        self.workspace_path = Path(os.getcwd())

    def run(self, *args: str, expect_fail: bool = False) -> None:
        try:
            tsrc.cli.main.main(args=args)
            rc = 0
        except SystemExit as e:
            rc = e.code
        if expect_fail and rc == 0:
            assert False, "should have failed"
        if rc != 0 and not expect_fail:
            raise SystemExit(rc)


@pytest.fixture
def tsrc_cli(workspace_path: Path, monkeypatch: Any) -> CLI:
    monkeypatch.chdir(workspace_path)
    res = CLI()
    return res



""" Entry point for tsrc push """

import argparse
import sys
from typing import Optional

import github3
from github3 import GitHub
from github3.pulls import PullRequest
import ui

import tsrc
import tsrc.github
from tsrc.cli.push import RepositoryInfo


class PushAction(tsrc.cli.push.PushAction):
    def __init__(self, repository_info: RepositoryInfo, args:
                 argparse.Namespace, github_api: Optional[GitHub] = None) -> None:
        super().__init__(repository_info, args)
        self.github_api = github_api
        self.repository = None
        self.pull_request = None

    def setup_service(self) -> None:
        if not self.github_api:
            self.github_api = tsrc.github.login()
        assert self.project_name
        owner, name = self.project_name.split("/")
        self.repository = self.github_api.repository(owner, name)

    def post_push(self) -> None:
        self.pull_request = self.ensure_pull_request()
        assert self.pull_request
        if self.args.close:
            ui.info_2("Closing merge request #%s" % self.pull_request.number)
            self.pull_request.close()
            return
        params = dict()
        if self.requested_target_branch:
            params["base"] = self.requested_target_branch

        if self.requested_title:
            params["title"] = self.requested_title

        self.pull_request.update(**params)

        if self.requested_reviewers:
            message = ["Requesting review from", ", ".join(self.requested_reviewers)]
            ui.info_2(*message)
            tsrc.github.request_reviewers(
                self.repository,
                self.pull_request.number,
                self.requested_reviewers
            )

        if self.requested_assignee:
            ui.info_2("Assigning to", self.requested_assignee)
            self.assign_pull_request()

        if self.args.merge:
            self.merge_pull_request()

        ui.info(ui.green, "::",
                ui.reset, "See pull request at", self.pull_request.html_url)

    def find_opened_pull_request(self) -> Optional[PullRequest]:
        assert self.repository
        for pull_request in self.repository.pull_requests():
            if pull_request.head.ref == self.remote_branch:
                if pull_request.state == "open":
                    return pull_request
        return None

    def create_pull_request(self) -> PullRequest:
        assert self.repository
        ui.info_2("Creating pull request", ui.ellipsis, end="")
        title = self.requested_title or self.remote_branch
        if self.requested_target_branch:
            target_branch = self.requested_target_branch
        else:
            target_branch = self.repository.default_branch
        try:
            pull_request = self.repository.create_pull(
                title,
                target_branch,
                self.remote_branch
            )
            ui.info("done", ui.check)
        except github3.GitHubError as github_error:
            ui.info()
            ui.error(ui.red, "\nCould not create pull request")
            for error in github_error.errors:
                ui.info(ui.red, error["message"])
            sys.exit(1)
        return pull_request

    def merge_pull_request(self) -> None:
        assert self.pull_request
        ui.info_2("Merging #", self.pull_request.number)
        self.pull_request.merge()

    def ensure_pull_request(self) -> PullRequest:
        pull_request = self.find_opened_pull_request()
        if pull_request:
            ui.info_2("Found existing pull request: #%s" % pull_request.number)
            return pull_request
        else:
            return self.create_pull_request()

    def assign_pull_request(self) -> None:
        assert self.pull_request
        issue = self.repository.issue(self.pull_request.number)
        issue.assign(self.requested_assignee)



from tsrc.test.helpers.cli import CLI
from ui.tests.conftest import message_recorder


def test_version(tsrc_cli: CLI, message_recorder: message_recorder) -> None:
    tsrc_cli.run("version")
    assert message_recorder.find("version")



""" Common tools """


from .config import parse_config_file  # noqa
from .config import Config, parse_tsrc_config, dump_tsrc_config, get_tsrc_config_path # noqa
from .errors import Error, InvalidConfig  # noqa
from .executor import Task, run_sequence, ExecutorFailed # noqa
from .groups import GroupList, Group  # noqa
from .groups import GroupNotFound, UnknownElement as UnknownGroupElement  # noqa
from .repo import Repo, Remote  # noqa
from .manifest import Manifest # noqa
from .manifest import load as load_manifest # noqa
from .workspace import Workspace # noqa



from typing import Any, Dict, Optional, List  # noqa
import argparse

import attr
from path import Path
import schema
import ruamel.yaml

import tsrc.config

MANIFEST_CONFIG_SCHEMA = schema.Schema({
    "url": str,
    schema.Optional("branch"): str,
    schema.Optional("tag"): str,
    schema.Optional("groups"): [str],
    schema.Optional("shallow"): bool,
})


@attr.s
class ManifestConfig:
    url = attr.ib(default=None)  # type: str
    branch = attr.ib(default="master")  # type: str
    tag = attr.ib(default=None)  # type: Optional[str]
    shallow = attr.ib(default=False)  # type: bool
    groups = attr.ib(default=list())  # type: List[str]

    @classmethod
    def from_dict(cls: "ManifestConfig", as_dict: dict) -> "ManifestConfig":
        res = ManifestConfig()
        res.url = as_dict["url"]
        res.branch = as_dict.get("branch", "master")
        res.tag = as_dict.get("tag")
        res.shallow = as_dict.get("shallow", False)
        res.groups = as_dict.get("groups") or list()
        return res

    @classmethod
    def from_args(cls: "ManifestConfig", args: argparse.Namespace) -> "ManifestConfig":
        as_dict = vars(args)  # type: dict
        return cls.from_dict(as_dict)

    @classmethod
    def from_file(cls: "ManifestConfig", cfg_path: Path) -> "ManifestConfig":
        as_dict = tsrc.config.parse_config_file(cfg_path, MANIFEST_CONFIG_SCHEMA)  # type: dict
        return cls.from_dict(as_dict)

    def save_to_file(self, cfg_path: Path) -> None:
        as_dict = self.as_dict()
        with cfg_path.open("w") as fp:
            ruamel.yaml.dump(as_dict, fp)

    def as_dict(self) -> Dict[str, Any]:
        res = dict()  # type: Dict[str, Any]
        res["url"] = self.url
        res["branch"] = self.branch
        if self.tag:
            res["tag"] = self.tag
        if self.groups:
            res["groups"] = self.groups
        res["shallow"] = self.shallow
        return res



import sys
from setuptools import setup, find_packages

if sys.version_info.major < 3:
    sys.exit("Error: Please upgrade to Python3")


def get_long_description():
    with open("README.rst") as fp:
        return fp.read()


setup(
    name="tsrc",
    version="0.6.1",
    description="Manage multiple repositories",
    long_description=get_long_description(),
    url="https://github.com/SuperTanker/tsrc",
    author="Kontrol SAS",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "attrs",
        "colored_traceback",
        "colorama",
        "github3.py >= 1.0",
        "path.py",
        "python-cli-ui",
        "python-gitlab",
        "pyparsing",
        "requests",
        "ruamel.yaml",
        "schema",
        "tabulate",
        "unidecode",
        "xdg",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    entry_points={"console_scripts": ["tsrc = tsrc.cli.main:main"]},
)



""" Entry point for `tsrc init` """
import argparse
import os

from path import Path
import ui

import tsrc
import tsrc.workspace.manifest_config


def main(args: argparse.Namespace) -> None:
    workspace_path = args.workspace_path or os.getcwd()
    workspace = tsrc.Workspace(Path(workspace_path))
    ui.info_1("Configuring workspace in", ui.bold, workspace_path)
    manifest_config = tsrc.workspace.ManifestConfig.from_args(args)
    workspace.configure_manifest(manifest_config)
    workspace.load_manifest()
    workspace.clone_missing()
    workspace.set_remotes()
    workspace.copy_files()
    ui.info("Done", ui.check)
